@@ -8,7 +8,7 @@ import tornado
 # local lib
 from jupyterlab_git.git import Git
 
-from .testutils import FakeContentManager
+from .testutils import FakeContentManager, maybe_future
 
 
 def test_is_remote_branch():
@@ -33,7 +33,7 @@ async def test_get_current_branch_success():
 
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future((0, "feature-foo", ""))
+        mock_execute.return_value = maybe_future((0, "feature-foo", ""))
 
         # When
         actual_response = await (
@@ -44,7 +44,7 @@ async def test_get_current_branch_success():
 
         # Then
         mock_execute.assert_called_once_with(
-            ["git", "symbolic-ref", "HEAD"], cwd=os.path.join("/bin", "test_curr_path")
+            ["git", "symbolic-ref", "--short", "HEAD"], cwd=os.path.join("/bin", "test_curr_path")
         )
         assert "feature-foo" == actual_response
 
@@ -59,10 +59,10 @@ async def test_checkout_branch_noref_success():
 
     with patch("jupyterlab_git.git.execute") as mock_execute:
         with patch.object(
-            Git, "_get_branch_reference", return_value=tornado.gen.maybe_future(None)
+            Git, "_get_branch_reference", return_value=maybe_future(None)
         ) as mock__get_branch_reference:
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -93,10 +93,10 @@ async def test_checkout_branch_noref_failure():
     rc = 1
     with patch("jupyterlab_git.git.execute") as mock_execute:
         with patch.object(
-            Git, "_get_branch_reference", return_value=tornado.gen.maybe_future(None)
+            Git, "_get_branch_reference", return_value=maybe_future(None)
         ) as mock__get_branch_reference:
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -132,10 +132,10 @@ async def test_checkout_branch_remoteref_success():
         with patch.object(
             Git,
             "_get_branch_reference",
-            return_value=tornado.gen.maybe_future("refs/remotes/remote_branch"),
+            return_value=maybe_future("refs/remotes/remote_branch"),
         ) as mock__get_branch_reference:
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -169,10 +169,10 @@ async def test_checkout_branch_headsref_failure():
         with patch.object(
             Git,
             "_get_branch_reference",
-            return_value=tornado.gen.maybe_future("refs/heads/local_branch"),
+            return_value=maybe_future("refs/heads/local_branch"),
         ) as mock__get_branch_reference:
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -206,10 +206,10 @@ async def test_checkout_branch_headsref_success():
         with patch.object(
             Git,
             "_get_branch_reference",
-            return_value=tornado.gen.maybe_future("refs/heads/local_branch"),
+            return_value=maybe_future("refs/heads/local_branch"),
         ):
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -239,10 +239,10 @@ async def test_checkout_branch_remoteref_failure():
         with patch.object(
             Git,
             "_get_branch_reference",
-            return_value=tornado.gen.maybe_future("refs/remotes/remote_branch"),
+            return_value=maybe_future("refs/remotes/remote_branch"),
         ):
             # Given
-            mock_execute.return_value = tornado.gen.maybe_future(
+            mock_execute.return_value = maybe_future(
                 (rc, stdout_message, stderr_message)
             )
 
@@ -272,7 +272,7 @@ async def test_get_branch_reference_success():
         branch = "test-branch"
         reference = "refs/remotes/origin/test_branch"
 
-        mock_execute.return_value = tornado.gen.maybe_future((0, reference, ""))
+        mock_execute.return_value = maybe_future((0, reference, ""))
 
         # When
         actual_response = await Git(FakeContentManager("/bin"))._get_branch_reference(
@@ -294,7 +294,7 @@ async def test_get_branch_reference_failure():
         branch = "test-branch"
         reference = "test-branch"
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (
                 128,
                 reference,
@@ -321,7 +321,7 @@ async def test_get_branch_reference_failure():
 async def test_get_current_branch_failure():
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (
                 128,
                 "",
@@ -337,11 +337,11 @@ async def test_get_current_branch_failure():
 
         # Then
         mock_execute.assert_called_once_with(
-            ["git", "symbolic-ref", "HEAD"], cwd=os.path.join("/bin", "test_curr_path")
+            ["git", "symbolic-ref", "--short", "HEAD"], cwd=os.path.join("/bin", "test_curr_path")
         )
         assert (
             "Error [fatal: Not a git repository (or any of the parent directories): .git] "
-            "occurred while executing [git symbolic-ref HEAD] command to get current branch."
+            "occurred while executing [git symbolic-ref --short HEAD] command to get current branch."
             == str(error.value)
         )
 
@@ -356,7 +356,7 @@ async def test_get_current_branch_detached_success():
             "  remotes/origin/feature-foo",
             "  remotes/origin/HEAD",
         ]
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (0, "\n".join(process_output), "")
         )
 
@@ -376,7 +376,7 @@ async def test_get_current_branch_detached_success():
 async def test_get_current_branch_detached_failure():
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (
                 128,
                 "",
@@ -403,17 +403,20 @@ async def test_get_current_branch_detached_failure():
 
 @pytest.mark.asyncio
 @pytest.mark.parametrize(
-    "branch,upstream",
+    "branch,upstream,remotename",
     [
-        ("feature-foo", "origin/master"),
-        ("master", "origin/master"),
-        ("feature-bar", "feature-foo"),
+        ("feature-foo", "master", "origin/withslash"),
+        ("master", "master", "origin"),
+        ("feature/bar", "feature-foo", ""),
     ],
 )
-async def test_get_upstream_branch_success(branch, upstream):
+async def test_get_upstream_branch_success(branch, upstream, remotename):
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future((0, upstream, ""))
+        mock_execute.side_effect = [
+            maybe_future((0, remotename + '/' + upstream, '')), 
+            maybe_future((0, remotename, ''))
+            ]
 
         # When
         actual_response = await Git(FakeContentManager("/bin")).get_upstream_branch(
@@ -421,11 +424,21 @@ async def test_get_upstream_branch_success(branch, upstream):
         )
 
         # Then
-        mock_execute.assert_called_once_with(
-            ["git", "rev-parse", "--abbrev-ref", "{}@{{upstream}}".format(branch)],
-            cwd=os.path.join("/bin", "test_curr_path"),
+        mock_execute.assert_has_calls(
+            [
+                call(
+                    ["git", "rev-parse", "--abbrev-ref", "{}@{{upstream}}".format(branch)],
+                    cwd=os.path.join("/bin", "test_curr_path"),
+                ),
+                call(
+                    ['git', 'config', '--local', 'branch.{}.remote'.format(branch)],
+                     cwd='/bin/test_curr_path',
+                ),
+
+            ],
+            any_order=False,
         )
-        assert upstream == actual_response
+        assert {'code': 0, 'remote_branch': upstream, 'remote_short_name': remotename} == actual_response
 
 
 @pytest.mark.asyncio
@@ -451,20 +464,15 @@ async def test_get_upstream_branch_success(branch, upstream):
 async def test_get_upstream_branch_failure(outputs, message):
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future(outputs)
+        mock_execute.return_value = maybe_future(outputs)
 
         # When
-        if message:
-            with pytest.raises(Exception) as error:
-                await Git(FakeContentManager("/bin")).get_upstream_branch(
-                    current_path="test_curr_path", branch_name="blah"
-                )
-            assert message == str(error.value)
-        else:
-            response = await Git(FakeContentManager("/bin")).get_upstream_branch(
-                current_path="test_curr_path", branch_name="blah"
-            )
-            assert response is None
+        response = await Git(FakeContentManager("/bin")).get_upstream_branch(
+            current_path="test_curr_path", branch_name="blah"
+        )
+        expected = {'code': 128, 'command': 'git rev-parse --abbrev-ref blah@{upstream}', 'message': outputs[2]}
+
+        assert response == expected
 
         # Then
         mock_execute.assert_has_calls(
@@ -482,7 +490,7 @@ async def test_get_upstream_branch_failure(outputs, message):
 async def test_get_tag_success():
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future((0, "v0.3.0", ""))
+        mock_execute.return_value = maybe_future((0, "v0.3.0", ""))
 
         # When
         actual_response = await Git(FakeContentManager("/bin"))._get_tag(
@@ -503,8 +511,8 @@ async def test_get_tag_failure():
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((128, "", "fatal: Not a valid object name blah")),
-            tornado.gen.maybe_future(
+            maybe_future((128, "", "fatal: Not a valid object name blah")),
+            maybe_future(
                 (
                     128,
                     "",
@@ -557,7 +565,7 @@ async def test_get_tag_failure():
 async def test_no_tags():
     with patch("jupyterlab_git.git.execute") as mock_execute:
         # Given
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (128, "", "fatal: No names found, cannot describe anything.\n")
         )
 
@@ -590,9 +598,9 @@ async def test_branch_success():
 
         mock_execute.side_effect = [
             # Response for get all refs/heads
-            tornado.gen.maybe_future((0, "\n".join(process_output_heads), "")),
+            maybe_future((0, "\n".join(process_output_heads), "")),
             # Response for get all refs/remotes
-            tornado.gen.maybe_future((0, "\n".join(process_output_remotes), "")),
+            maybe_future((0, "\n".join(process_output_remotes), "")),
         ]
 
         expected_response = {
@@ -694,7 +702,7 @@ async def test_branch_failure():
             "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)",
             "refs/heads/",
         ]
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (
                 128,
                 "",
@@ -738,15 +746,15 @@ async def test_branch_success_detached_head():
 
         mock_execute.side_effect = [
             # Response for get all refs/heads
-            tornado.gen.maybe_future((0, "\n".join(process_output_heads), "")),
+            maybe_future((0, "\n".join(process_output_heads), "")),
             # Response for get current branch
-            tornado.gen.maybe_future(
+            maybe_future(
                 (128, "", "fatal: ref HEAD is not a symbolic ref")
             ),
             # Response for get current branch detached
-            tornado.gen.maybe_future((0, "\n".join(detached_head_output), "")),
+            maybe_future((0, "\n".join(detached_head_output), "")),
             # Response for get all refs/remotes
-            tornado.gen.maybe_future((0, "\n".join(process_output_remotes), "")),
+            maybe_future((0, "\n".join(process_output_remotes), "")),
         ]
 
         expected_response = {
@@ -807,7 +815,7 @@ async def test_branch_success_detached_head():
                 ),
                 # call to get current branch
                 call(
-                    ["git", "symbolic-ref", "HEAD"],
+                    ["git", "symbolic-ref", "--short", "HEAD"],
                     cwd=os.path.join("/bin", "test_curr_path"),
                 ),
                 # call to get current branch name given a detached head
