@@ -2,8 +2,10 @@ import json
 import os
 from unittest.mock import ANY, Mock, call, patch
 
+import requests
 import tornado
 
+from jupyterlab_git.git import Git
 from jupyterlab_git.handlers import (
     GitAllHistoryHandler,
     GitBranchHandler,
@@ -13,7 +15,7 @@ from jupyterlab_git.handlers import (
     setup_handlers,
 )
 
-from .testutils import ServerTest, assert_http_error
+from .testutils import ServerTest, assert_http_error, maybe_future
 
 
 def test_mapping_added():
@@ -25,7 +27,7 @@ def test_mapping_added():
 
 
 class TestAllHistory(ServerTest):
-    @patch("jupyterlab_git.handlers.GitAllHistoryHandler.git")
+    @patch("jupyterlab_git.handlers.GitAllHistoryHandler.git", spec=Git)
     def test_all_history_handler_localbranch(self, mock_git):
         # Given
         show_top_level = {"code": 0, "foo": "top_level"}
@@ -33,10 +35,10 @@ class TestAllHistory(ServerTest):
         log = "log_foo"
         status = "status_foo"
 
-        mock_git.show_top_level.return_value = tornado.gen.maybe_future(show_top_level)
-        mock_git.branch.return_value = tornado.gen.maybe_future(branch)
-        mock_git.log.return_value = tornado.gen.maybe_future(log)
-        mock_git.status.return_value = tornado.gen.maybe_future(status)
+        mock_git.show_top_level.return_value = maybe_future(show_top_level)
+        mock_git.branch.return_value = maybe_future(branch)
+        mock_git.log.return_value = maybe_future(log)
+        mock_git.status.return_value = maybe_future(status)
 
         # When
         body = {"current_path": "test_path", "history_count": 25}
@@ -62,7 +64,7 @@ class TestAllHistory(ServerTest):
 
 
 class TestBranch(ServerTest):
-    @patch("jupyterlab_git.handlers.GitBranchHandler.git")
+    @patch("jupyterlab_git.handlers.GitBranchHandler.git", spec=Git)
     def test_branch_handler_localbranch(self, mock_git):
         # Given
         branch = {
@@ -111,7 +113,7 @@ class TestBranch(ServerTest):
             ],
         }
 
-        mock_git.branch.return_value = tornado.gen.maybe_future(branch)
+        mock_git.branch.return_value = maybe_future(branch)
 
         # When
         body = {"current_path": "test_path"}
@@ -126,11 +128,11 @@ class TestBranch(ServerTest):
 
 
 class TestLog(ServerTest):
-    @patch("jupyterlab_git.handlers.GitLogHandler.git")
+    @patch("jupyterlab_git.handlers.GitLogHandler.git", spec=Git)
     def test_log_handler(self, mock_git):
         # Given
         log = {"code": 0, "commits": []}
-        mock_git.log.return_value = tornado.gen.maybe_future(log)
+        mock_git.log.return_value = maybe_future(log)
 
         # When
         body = {"current_path": "test_path", "history_count": 20}
@@ -143,11 +145,11 @@ class TestLog(ServerTest):
         payload = response.json()
         assert payload == log
 
-    @patch("jupyterlab_git.handlers.GitLogHandler.git")
+    @patch("jupyterlab_git.handlers.GitLogHandler.git", spec=Git)
     def test_log_handler_no_history_count(self, mock_git):
         # Given
         log = {"code": 0, "commits": []}
-        mock_git.log.return_value = tornado.gen.maybe_future(log)
+        mock_git.log.return_value = maybe_future(log)
 
         # When
         body = {"current_path": "test_path"}
@@ -162,14 +164,18 @@ class TestLog(ServerTest):
 
 
 class TestPush(ServerTest):
-    @patch("jupyterlab_git.handlers.GitPushHandler.git")
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
     def test_push_handler_localbranch(self, mock_git):
         # Given
-        mock_git.get_current_branch.return_value = tornado.gen.maybe_future("foo")
-        mock_git.get_upstream_branch.return_value = tornado.gen.maybe_future(
-            "localbranch"
+        mock_git.get_current_branch.return_value = maybe_future("localbranch")
+        mock_git.get_upstream_branch.return_value = maybe_future(
+            {
+                "code": 0,
+                "remote_short_name": ".",
+                "remote_branch": "localbranch"
+            }
         )
-        mock_git.push.return_value = tornado.gen.maybe_future({"code": 0})
+        mock_git.push.return_value = maybe_future({"code": 0})
 
         # When
         body = {"current_path": "test_path"}
@@ -177,21 +183,22 @@ class TestPush(ServerTest):
 
         # Then
         mock_git.get_current_branch.assert_called_with("test_path")
-        mock_git.get_upstream_branch.assert_called_with("test_path", "foo")
-        mock_git.push.assert_called_with(".", "HEAD:localbranch", "test_path", None)
+        mock_git.get_upstream_branch.assert_called_with("test_path", "localbranch")
+        mock_git.push.assert_called_with(".", "HEAD:localbranch", "test_path", None, False)
 
         assert response.status_code == 200
         payload = response.json()
         assert payload == {"code": 0}
 
-    @patch("jupyterlab_git.handlers.GitPushHandler.git")
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
     def test_push_handler_remotebranch(self, mock_git):
         # Given
-        mock_git.get_current_branch.return_value = tornado.gen.maybe_future("foo")
-        mock_git.get_upstream_branch.return_value = tornado.gen.maybe_future(
-            "origin/remotebranch"
-        )
-        mock_git.push.return_value = tornado.gen.maybe_future({"code": 0})
+        mock_git.get_current_branch.return_value = maybe_future("foo/bar")
+        upstream = {"code": 0,
+                    "remote_short_name": "origin/something",
+                    "remote_branch": "remote-branch-name"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.push.return_value = maybe_future({"code": 0})
 
         # When
         body = {"current_path": "test_path"}
@@ -199,45 +206,213 @@ class TestPush(ServerTest):
 
         # Then
         mock_git.get_current_branch.assert_called_with("test_path")
-        mock_git.get_upstream_branch.assert_called_with("test_path", "foo")
+        mock_git.get_upstream_branch.assert_called_with("test_path", "foo/bar")
         mock_git.push.assert_called_with(
-            "origin", "HEAD:remotebranch", "test_path", None
+            "origin/something", "HEAD:remote-branch-name", "test_path", None, False
         )
 
         assert response.status_code == 200
         payload = response.json()
         assert payload == {"code": 0}
 
-    @patch("jupyterlab_git.handlers.GitPushHandler.git")
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
     def test_push_handler_noupstream(self, mock_git):
         # Given
-        mock_git.get_current_branch.return_value = tornado.gen.maybe_future("foo")
-        mock_git.get_upstream_branch.return_value = tornado.gen.maybe_future("")
-        mock_git.push.return_value = tornado.gen.maybe_future({"code": 0})
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {
+            "code": 128,
+            "command": "",
+            "message": "fatal: no upstream configured for branch 'foo'"
+        }
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": dict()})
+        mock_git.remote_show.return_value = maybe_future({})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
 
         # When
-        body = {"current_path": "test_path"}
-        response = self.tester.post(["push"], body=body)
+        body = {"current_path": path}
+        try:
+            self.tester.post(["push"], body=body)
+        except requests.exceptions.HTTPError as error:
+            response = error.response
 
         # Then
-        mock_git.get_current_branch.assert_called_with("test_path")
-        mock_git.get_upstream_branch.assert_called_with("test_path", "foo")
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_called_with(path)
+        mock_git.remote_show.assert_called_with(path)
         mock_git.push.assert_not_called()
 
-        assert response.status_code == 200
+        assert response.status_code == 500
         payload = response.json()
         assert payload == {
             "code": 128,
             "message": "fatal: The current branch foo has no upstream branch.",
+            "remotes": list()
         }
 
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
+    def test_push_handler_multipleupstream(self, mock_git):
+        # Given
+        remotes = ["origin", "upstream"]
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {"code": -1, "message": "oups"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": dict()})
+        mock_git.remote_show.return_value = maybe_future({"remotes": remotes})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
+
+        # When
+        body = {"current_path": path}
+        try:
+            self.tester.post(["push"], body=body)
+        except requests.exceptions.HTTPError as error:
+            response = error.response
+
+        # Then
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_called_with(path)
+        mock_git.remote_show.assert_called_with(path)
+        mock_git.push.assert_not_called()
+
+        assert response.status_code == 500
+        payload = response.json()
+        assert payload == {
+            "code": 128,
+            "message": "fatal: The current branch foo has no upstream branch.",
+            "remotes": remotes
+        }
+
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
+    def test_push_handler_noupstream_unique_remote(self, mock_git):
+        # Given
+        remote = "origin"
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {"code": -1, "message": "oups"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": dict()})
+        mock_git.remote_show.return_value = maybe_future({"remotes": [remote]})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
+
+        # When
+        body = {"current_path": path}
+        response = self.tester.post(["push"], body=body)
+
+        # Then
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_called_with(path)
+        mock_git.remote_show.assert_called_with(path)
+        mock_git.push.assert_called_with(remote, "foo", "test_path", None, set_upstream=True)
+
+        assert response.status_code == 200
+        payload = response.json()
+        assert payload == {"code": 0}
+
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
+    def test_push_handler_noupstream_pushdefault(self, mock_git):
+        # Given
+        remote = "rorigin"
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {"code": -1, "message": "oups"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": {"remote.pushdefault": remote}})
+        mock_git.remote_show.return_value = maybe_future({"remotes": [remote, "upstream"]})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
+
+        # When
+        body = {"current_path": path}
+        response = self.tester.post(["push"], body=body)
+
+        # Then
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_called_with(path)
+        mock_git.remote_show.assert_called_with(path)
+        mock_git.push.assert_called_with(remote, "foo", "test_path", None, set_upstream=True)
+
+        assert response.status_code == 200
+        payload = response.json()
+        assert payload == {"code": 0}
+
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
+    def test_push_handler_noupstream_pass_remote_nobranch(self, mock_git):
+        # Given
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {"code": -1, "message": "oups"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": dict()})
+        mock_git.remote_show.return_value = maybe_future({})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
+        remote = "online"
+
+        # When
+        body = {"current_path": path, "remote": remote}
+        response = self.tester.post(["push"], body=body)
+
+        # Then
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_not_called()
+        mock_git.remote_show.assert_not_called()
+        mock_git.push.assert_called_with(remote, "HEAD:foo", "test_path", None, True)
+
+        assert response.status_code == 200
+        payload = response.json()
+        assert payload == {"code": 0}
+
+    @patch("jupyterlab_git.handlers.GitPushHandler.git", spec=Git)
+    def test_push_handler_noupstream_pass_remote_branch(self, mock_git):
+        # Given
+        mock_git.get_current_branch.return_value = maybe_future("foo")
+        upstream = {"code": -1, "message": "oups"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+        mock_git.config.return_value = maybe_future({"options": dict()})
+        mock_git.remote_show.return_value = maybe_future({})
+        mock_git.push.return_value = maybe_future({"code": 0})
+
+        path = "test_path"
+        remote = "online"
+        remote_branch = "onfoo"
+
+        # When
+        body = {"current_path": path, "remote": "/".join((remote, remote_branch))}
+        response = self.tester.post(["push"], body=body)
+
+        # Then
+        mock_git.get_current_branch.assert_called_with(path)
+        mock_git.get_upstream_branch.assert_called_with(path, "foo")
+        mock_git.config.assert_not_called()
+        mock_git.remote_show.assert_not_called()
+        mock_git.push.assert_called_with(
+            remote, "HEAD:" + remote_branch, "test_path", None, True
+        )
+
+        assert response.status_code == 200
+        payload = response.json()
+        assert payload == {"code": 0}
+
 
 class TestUpstream(ServerTest):
-    @patch("jupyterlab_git.handlers.GitUpstreamHandler.git")
-    def test_upstream_handler_localbranch(self, mock_git):
+    @patch("jupyterlab_git.handlers.GitUpstreamHandler.git", spec=Git)
+    def test_upstream_handler_forward_slashes(self, mock_git):
         # Given
-        mock_git.get_current_branch.return_value = tornado.gen.maybe_future("foo")
-        mock_git.get_upstream_branch.return_value = tornado.gen.maybe_future("bar")
+        mock_git.get_current_branch.return_value = maybe_future("foo/bar")
+        upstream = {"code": 0,
+                    "remote_short_name": "origin/something",
+                    "remote_branch": "foo/bar"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
 
         # When
         body = {"current_path": "test_path"}
@@ -245,11 +420,32 @@ class TestUpstream(ServerTest):
 
         # Then
         mock_git.get_current_branch.assert_called_with("test_path")
-        mock_git.get_upstream_branch.assert_called_with("test_path", "foo")
+        mock_git.get_upstream_branch.assert_called_with("test_path", "foo/bar")
 
         assert response.status_code == 200
         payload = response.json()
-        assert payload == {"upstream": "bar"}
+        assert payload == upstream 
+
+    @patch("jupyterlab_git.handlers.GitUpstreamHandler.git", spec=Git)
+    def test_upstream_handler_localbranch(self, mock_git):
+        # Given
+        mock_git.get_current_branch.return_value = maybe_future("foo/bar")
+        upstream = {"code": 0,
+                    "remote_short_name": ".",
+                    "remote_branch": "foo/bar"}
+        mock_git.get_upstream_branch.return_value = maybe_future(upstream)
+
+        # When
+        body = {"current_path": "test_path"}
+        response = self.tester.post(["upstream"], body=body)
+
+        # Then
+        mock_git.get_current_branch.assert_called_with("test_path")
+        mock_git.get_upstream_branch.assert_called_with("test_path", "foo/bar")
+
+        assert response.status_code == 200
+        payload = response.json()
+        assert payload == upstream 
 
 
 class TestDiffContent(ServerTest):
@@ -261,10 +457,10 @@ class TestDiffContent(ServerTest):
         content = "dummy content file\nwith multiple lines"
 
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, "")),
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, ""))
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, "")),
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, ""))
         ]
 
         # When
@@ -303,9 +499,9 @@ class TestDiffContent(ServerTest):
         content = "dummy content file\nwith multiple lines"
 
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, "")),
-            tornado.gen.maybe_future((0, content, ""))
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, "")),
+            maybe_future((0, content, ""))
         ]
 
         dummy_file = os.path.join(self.notebook_dir, top_repo_path, filename)
@@ -344,10 +540,10 @@ class TestDiffContent(ServerTest):
         content = "dummy content file\nwith multiple lines"
 
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, "")),
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, ""))
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, "")),
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, ""))
         ]
 
         # When
@@ -386,10 +582,10 @@ class TestDiffContent(ServerTest):
         content = "dummy content file\nwith multiple lines"
 
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, "")),
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, ""))
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, "")),
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, ""))
         ]
 
         # When
@@ -409,7 +605,7 @@ class TestDiffContent(ServerTest):
         top_repo_path = "path/to/repo"
         filename = "my/file"
 
-        mock_execute.return_value = tornado.gen.maybe_future(
+        mock_execute.return_value = maybe_future(
             (
                 -1,
                 "",
@@ -440,7 +636,7 @@ class TestDiffContent(ServerTest):
         top_repo_path = "path/to/repo"
         filename = "my/file"
 
-        mock_execute.return_value = tornado.gen.maybe_future((0, "-\t-\t{}".format(filename), ""))
+        mock_execute.return_value = maybe_future((0, "-\t-\t{}".format(filename), ""))
 
         # When
         body = {
@@ -460,7 +656,7 @@ class TestDiffContent(ServerTest):
         top_repo_path = "path/to/repo"
         filename = "my/file"
 
-        mock_execute.return_value = tornado.gen.maybe_future((-1, "", "Dummy error"))
+        mock_execute.return_value = maybe_future((-1, "", "Dummy error"))
 
         # When
         body = {
@@ -482,9 +678,9 @@ class TestDiffContent(ServerTest):
         content = "dummy content file\nwith multiple lines"
 
         mock_execute.side_effect = [
-            tornado.gen.maybe_future((0, "1\t1\t{}".format(filename), "")),
-            tornado.gen.maybe_future((0, content, "")),
-            tornado.gen.maybe_future((0, content, ""))
+            maybe_future((0, "1\t1\t{}".format(filename), "")),
+            maybe_future((0, content, "")),
+            maybe_future((0, content, ""))
         ]
 
         # When
